/* code to allow nos to talk directly to bpqcode
 *
 * by paul h launspach n5faz
 * (c) 1993 FazCom
 * 4-13-93
 */
/* $Id: bpqsock.c 1.5 1994/03/23 06:38:13 fz-phl Exp $ */
/* $Log: bpqsock.c $
 * Revision 1.5  1994/03/23  06:38:13  fz-phl
 * initial revision
 *
 */

#include <errno.h>
#include "global.h"
#include "socket.h"
#include "usock.h"

static void autobind(struct usock *up);
static void s_brcall(struct bpq_cb *cb, int cnt);
static void s_btcall(struct bpq_cb *cb, int cnt);
static void s_bscall(struct bpq_cb *cb, int old, int new);


int
so_bpq(up, protocol)
struct usock *up;
int protocol;
{
    return 0;
}

int
so_bpq_listen(up, backlog)
struct usock *up;
int backlog;
{
    struct sockaddr_bq *local;
    struct bpq_cb *cb;
    int s;

    if(up->name == NULLCHAR)
        return -1;
    s = up->index;
    local = (struct sockaddr_bq *)up->name;
    up->cb.bpq = cb = open_bpq(local->bpq_addr,NULLCHAR,
      backlog ? AX_SERVER:AX_PASSIVE,
      s_brcall,s_btcall,s_bscall,s);
    if(cb == NULLBPQ)
        return -1;
    if(!(cb->appl = local->bpq_appl))  /* appl must be non-zero for  */
        return -1;                     /* the BPQ_SET to work  */
    cb->applflag = local->bpq_flag;
    /* tell bpq to allow connections ... */
    bpq_int(BPQ_SET,cb->stream,cb->appl <<8 | cb->applflag,NULLCHAR);
    return 0;
}

int
so_bpq_conn(up)
struct usock *up;
{
    struct sockaddr_bq *local,*remote;
    struct bpq_cb *bp;
    int s;

    s = up->index;
    if(up->name == NULLCHAR)
        autobind(up);
    local = (struct sockaddr_bq *)up->name;
    remote = (struct sockaddr_bq *)up->peername;
    up->cb.bpq = open_bpq(local->bpq_addr,remote->bpq_addr,AX_ACTIVE,
      s_brcall,s_btcall,s_bscall,s);

    while((bp = up->cb.bpq) != NULLBPQ && bp->state != BPQ_ST_CON){
        if(up->noblock){
            errno = EWOULDBLOCK;
            return -1;
        } else if((errno = pwait(up)) != 0){
            return -1;
        }
    }
    if(bp == NULLBPQ){
        free(up->peername);
        up->peername = NULLCHAR;
        errno = ECONNREFUSED;
        return -1;
    }
    return 0;
}

int
so_bpq_recv(up,bpp,from,fromlen)
struct usock *up;
struct mbuf **bpp;
char *from;
int *fromlen;
{
    struct bpq_cb *cb;

    while((cb = up->cb.bpq) != NULLBPQ
     && (*bpp = recv_bpq(cb,(uint16)0)) == NULLBUF){
		if(up->noblock){
			errno = EWOULDBLOCK;
			return -1;
		} else if((errno = pwait(up)) != 0){
			return -1;
		}
	}
    if(cb == NULLBPQ){
		/* Connection went away */
		errno = ENOTCONN;
		return -1;
	}
	return (*bpp)->cnt;

}

int
so_bpq_send(up,bp,to)
struct usock *up;
struct mbuf *bp;
char *to;
{
    struct bpq_cb *cb;
    union sp sp;

    if((cb = up->cb.bpq) == NULLBPQ) {
        free_p(bp);
        errno = ENOTCONN;
        return -1;
    }
    send_bpq(cb,bp);
    sp.p = up->peername;
    memcpy(sp.bq->bpq_addr,cb->remote,AXBUF);

    while((cb = up->cb.bpq) != NULLBPQ &&
      cb->unack >= cb->maxframe){
        if(up->noblock){
            errno = EWOULDBLOCK;
            return -1;
        } else if((errno = pwait(up)) != 0){
            return -1;
        }
    }
    if(cb == NULLBPQ){
        errno = EBADF;
        return -1;
    }
    return 0;
}

int
so_bpq_qlen(up,rtx)
struct usock *up;
int rtx;
{
	int len;

	switch(rtx){
	case 0:
        len = len_p(up->cb.bpq->rxq);
		break;
	case 1:	/* Number of packets, not bytes */
        len = len_q(up->cb.bpq->txq) + up->cb.bpq->unack;
		break;
	}
	return len;

}

int
so_bpq_shut(up,how)
struct usock *up;
int how;
{
	switch(how){
    case 0: /* return them to the node */
	case 1:	/* Attempt regular disconnect */
        disc_bpq(up->cb.bpq,how);
		break;
	case 2: /* Blow it away */
        disc_bpq(up->cb.bpq,1);
        up->cb.bpq = NULLBPQ;
		break;
	}
	return 0;

}

int
so_bpq_close(up)
struct usock *up;
{
    if(up->cb.bpq != NULLBPQ){
        up->cb.bpq->user = -1;
        disc_bpq(up->cb.bpq,0);
        up->cb.bpq = NULLBPQ;
    }
    return 0;
}

static void
autobind(up)
struct usock *up;
{
    struct sockaddr_bq local;
    int s;

    s = up->index;
    local.bpq_family = AF_BPQ;
    pax25(local.bpq_addr,Mycall);
    local.bpq_appl = BPQ_NONE;
    local.bpq_flag = BPQ_NONE;
    bind(s,(char *)&local,sizeof(struct sockaddr_bq));
}

static void
s_brcall(cb,cnt)
struct bpq_cb *cb;
int cnt;
{
    psignal(itop(cb->user),1);
    pwait(NULL);
}

static void
s_btcall(cb,cnt)
struct bpq_cb *cb;
int cnt;
{
    psignal(itop(cb->user),1);
    pwait(NULL);
}

static void
s_bscall(cb,old,new)
struct bpq_cb *cb;
uint16 old;
uint16 new;
{
    int s,ns;
    struct usock *up,*nup,*oup;
    union sp sp;

    s = cb->user;
    oup = up = itop(s);

    switch(new){
    case BPQ_ST_DIS:
        if(up != NULLUSOCK){
            up->cb.bpq = NULLBPQ;
        }
        break;
    case BPQ_ST_CON:
        if(cb->mode == AX_SERVER){
            ns = socket(AF_BPQ,SOCK_STREAM,0);
			nup = itop(ns);
			cb->user = ns;
            cb->mode = AX_PASSIVE;
            nup->cb.bpq = cb;
			/* Allocate new memory for the name areas */
            nup->name = mallocw(sizeof(struct sockaddr_bq));
            nup->peername = mallocw(sizeof(struct sockaddr_bq));
			/* Store the new socket # in the old one */
			up->rdysock = ns;
			up = nup;
			s = ns;
		} else {
			/* Allocate space for the peer's name */
            up->peername = mallocw(sizeof(struct sockaddr_bq));
			/* Store the old socket # in the old socket */
			up->rdysock = s;
		}
		/* Load the addresses. Memory for the name has already
		 * been allocated, either above or in the original bind.
		 */
		sp.p = up->name;
        sp.bq->bpq_family = AF_BPQ;
        memcpy(sp.bq->bpq_addr,cb->local,AXBUF);
        up->namelen = sizeof(struct sockaddr_bq);

		sp.p = up->peername;
        sp.bq->bpq_family = AF_BPQ;
        memcpy(sp.bq->bpq_addr,cb->remote,AXBUF);
        up->peernamelen = sizeof(struct sockaddr_bq);

		/* Wake up the guy accepting it, and let him run */
		psignal(oup,1);
		pwait(NULL);
        break;
    default:
        break;
	}
    psignal(up,0);
}

char *
bpqpsocket(p)
struct sockaddr *p;
{
    static char buf[11];
    struct sockaddr_bq *sp = (struct sockaddr_bq *)p;
    sprintf(buf,"%10.10s",sp->bpq_addr);
    return buf;
}

char *
bpqstate(up)
struct usock *up;
{
    return Bpqstates[up->cb.bpq->state];
}

int
so_bpq_stat(up)
struct usock *up;
{
    printf("&BPQCB   Owner            Sock St Num Recv Sent  Rxq  Txq Appl Tot Curr\n");
    st_bpq_cb(up->cb.bpq);
    printf(" St Num State       Type/Pt User      Recv Unak Maxf Pacl\n");
    st_bpq(up->cb.bpq->stream);
    return 0;
}

