/* code to allow nos to talk directly to bpqcode
 *
 * by paul h launspach n5faz
 * (c) 1993 FazCom
 * 4-13-93
 */
/* $Id: bpquser.c 1.5 1994/03/23 06:38:13 fz-phl Exp $ */
/* $Log: bpquser.c $
 * Revision 1.5  1994/03/23  06:38:13  fz-phl
 * initial revision
 *
 */

#include "global.h"
#include "mbuf.h"
#include "ax25.h"
#include "bpq.h"

/* send data to a streem */
int
send_bpq(cb,bp)
struct bpq_cb *cb;
struct mbuf *bp;
{
    struct mbuf *bp1;
    uint16 offset,len,size;
    char tmp[11];

    if(cb == NULLBPQ || bp == NULLBUF){
        free_p(bp);
        return -1;
    }
    /* paclen (and maxf) can change while a connection is active */
    cb->maxframe = bpq_int(BPQ_MAXFR,cb->stream,0,tmp) + Bpqinfo.window;
    cb->paclen = max(bpq_int(BPQ_PACLEN,cb->stream,0,tmp),21);
    memcpy(cb->remote,tmp,AXBUF);
    offset = 0;
    len = len_p(bp);
    while(len != 0){
        size = min(len,cb->paclen);
        dup_p(&bp1,bp,offset,size);
        len -=size;
        offset += size;
        enqueue(&cb->txq,bp1);
    }
    free_p(bp);
    return bpq_out(cb);
}

/* recieve from a stream */
struct mbuf *
recv_bpq(cb,cnt)
struct bpq_cb *cb;
uint16 cnt;
{
    struct mbuf *bp;

    if(cb->rxq == NULLBUF)
        return NULLBUF;

    if(cnt == 0){
        bp = cb->rxq;
        cb->rxq = NULLBUF;
    } else {
        bp = ambufw(cnt);
        bp->cnt = pullup(&cb->rxq,bp->data,cnt);
    }
    return bp;
}

/* low level send data to a streem */
int
bpq_out(cb)
struct bpq_cb *cb;
{
    struct mbuf *bp;
    int sent = 0;
    uint16 bufs;


    /* check for free buffers in bpqcode. leave some fudge for rx frames */
    bufs = bpq_int(BPQ_BUF,1,0,NULLCHAR);
    Bpqinfo.minfree = min(bufs,Bpqinfo.minfree);
    bufs -= (cb->maxframe + 20);

    cb->unack = bpq_int(BPQ_UNACK,cb->stream,0,NULLCHAR);

    /* dont over stuff bpqcode */
    while(bufs > sent && cb->maxframe > cb->unack
      && (bp = dequeue(&cb->txq)) != NULLBUF){

        bpq_int(BPQ_SEND,cb->stream,bp->cnt,bp->data);
        free_p(bp);
        cb->unack = bpq_int(BPQ_UNACK,cb->stream,0,NULLCHAR);
        sent ++;
    }
    /* un block us ... */
    if(cb->maxframe > cb->unack)
        if(cb->t_upcall != NULLVFP)
            (*cb->t_upcall)(cb,len_q(cb->txq));

    cb->send += sent;
    Bpqinfo.send += sent;
    return sent;
}


/* Verify that cb points to a valid bpq control block */
int
bpqval(cb)
struct bpq_cb *cb;
{
    register struct bpq_cb *cb1;

    if(cb == NULLBPQ)
		return 0;	/* Null pointer can't be valid */
    for(cb1 = Bpqinfo.cb;cb1 != NULLBPQ; cb1 = cb1->next)
        if(cb1 == cb)
			return 1;
	return 0;
}

/* return bpq_cb for connection on stream  */
struct bpq_cb *
find_bpq(stream)
int stream;
{
    register struct bpq_cb *cb;

    for(cb=Bpqinfo.cb; cb != NULLBPQ; cb = cb->next)
        if(stream == cb->stream)
            break;

    return cb;
}

/* open */
struct bpq_cb *
open_bpq(local,remote,mode,r_upcall,t_upcall,s_upcall,user)
char *local;
char *remote;
int mode;
void (*r_upcall)();
void (*t_upcall)();
void (*s_upcall)();
int user;
{
    struct bpq_cb *cb;

    if((cb = cr_bpq()) == NULLBPQ)
        return NULLBPQ;

    if(remote == NULLCHAR)
        remote = "SWITCH   ";

    memcpy(cb->local,local,AXBUF);
    memcpy(cb->remote,remote,AXBUF);
    cb->r_upcall = r_upcall;
    cb->t_upcall = t_upcall;
    cb->s_upcall = s_upcall;
    cb->user = user;
    cb->mode = mode;

    switch(mode){
        case AX_SERVER:
        case AX_PASSIVE:
            cb->state = BPQ_ST_LIS;
            return cb;
        case AX_ACTIVE:
            break;
    }
    /* connect to node */
    bpq_int(BPQ_CONNECT,cb->stream,BPQ_CONN,NULLCHAR);

    return cb;
}

/* disconnect */
void
disc_bpq(cb,ret)
struct bpq_cb *cb;
int ret;                /* 0 = return user to node, 1 = disc em */
{
    /* stop any new connection on this stream */
    bpq_int(BPQ_SET,cb->stream,0,NULLCHAR);

    if(cb->state == BPQ_ST_LIS){
        del_bpq(cb);
        return;
    }

    /* and disconnect it */
    bpq_int(BPQ_CONNECT,cb->stream,ret ? BPQ_DISC : BPQ_RET,NULLCHAR);
    psignal(Bpqinfo.rxproc,0);

    return;
}

void
reset_bpq(cb)
struct bpq_cb *cb;
{
    if(cb->s_upcall != NULLVFP)
        (*cb->s_upcall)(cb,cb->state,BPQ_ST_DIS);

    disc_bpq(cb,1);
}

/* kick a bpq connection ... does nothing  */
int
kick_bpq(cb)
struct bpq_cb *cb;
{
    return 0;
}

